// Copyright (c) 2021 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include <stdint.h>
#include <malloc.h>
#include "tile.h"
#include "./../util/math.h"
#include "../display/shader.h"
#include "../display/primitive.h"
#include "../display/texture.h"

#define CHUNK_WIDTH 3
#define CHUNK_HEIGHT CHUNK_WIDTH
#define CHUNK_DEPTH CHUNK_HEIGHT

typedef struct {
  /** Absolute X Y Z chunklist coordinates */
  int32_t x, y, z;
  
  /** Tiles stored within the chunk. */
  tile_t *tiles;

  /** Primitive used for rendering the chunk */
  primitive_t *primitive;
} chunk_t;

typedef struct {
  /** Dimensions of the chunk list */
  int32_t width, height, depth;

  /** Position of index 0 */
  int32_t x, y, z;

  /** Count of chunks in the internal array */
  int32_t count;

  /** Count of tiles within each chunk */
  int32_t tileCount;

  /** Current order of each chunk in the list. */
  chunk_t **chunkList;

  /** Chunk list tile map. */
  tilemap_t *tilemap;

  /** The actual chunks in the list */
  chunk_t *chunks;
} chunklist_t;

/**
 * Creates a new chunk list. Chunk lists are managed to be memory efficient by
 * only keeping a small fraction of the overall size loaded at any given time.
 *
 * @param width The width (x axis) of chunks to keep loaded.
 * @param height The height (y axis) of chunks to keep loaded.
 * @param depth The depth (z axis) of chunks to keep loaded.
 * @param tilemap Tilemap of the chunk list.
 * @return A new chunk list.
 */
chunklist_t * chunkListCreate(int32_t width, int32_t height, int32_t depth,
  tilemap_t *tilemap
);

/**
 * Disposes and frees a previously created list. This does not free the chunks
 * themselves.

 * @param list The list to free.
 */
void chunkListDispose(chunklist_t *list);

/**
 * Shift the chunk list along a set of axis (in absolute space).
 * 
 * @param list List to shift.
 * @param x X movement to shift chunk along.
 * @param y Y movement to shift chunk along.
 * @param z Z movement to shift chunk along.
 */
void chunkListShift(chunklist_t *list, int32_t x, int32_t y, int32_t z);

/**
 * Align the chunk list (in absolute space).
 * 
 * @param list List to align.
 * @param x X movement to shift chunk along.
 * @param y Y movement to shift chunk along.
 * @param z Z movement to shift chunk along.
 */
void chunkListAlign(chunklist_t *list, int32_t x, int32_t y, int32_t z);

/**
 * Renders a given chunk list.
 * 
 * @param list The list to render.
 * @param shader The shader to use.
 */
void chunkListRender(chunklist_t *list, shader_t *shader);

/**
 * Loads a given chunk.
 * 
 * @param list List to load the chunk for.
 * @param chunk The chunk to load.
 * @param x X of the chunk.
 * @param y Y of the chunk.
 * @param z Z of the chunk.
 */
void chunkListChunkLoad(chunklist_t *list, chunk_t *chunk,
  int32_t x, int32_t y, int32_t z
);

/**
 * Unload a given chunk.
 * 
 * @param list List that the chunk belongs to.
 * @param chunk Chunk to unload.
 */
void chunkListChunkUnload(chunklist_t *list, chunk_t *chunk);

/**
 * Rendedrs a given chunk tile.
 * 
 * @param list The chunk list that the chunk belongs to.
 * @param chunk The chunk to render against.
 * @param tile The tile to render.
 * @param tileDef The tile's definition information.
 * @param x The X Coordinate of the tile (chunk space).
 * @param y The Y Coordinate of the tile (chunk space).
 * @param z The Z Coordinate of the tile (chunk space).
 */
void chunkListTileRender(chunklist_t *list, chunk_t *chunk,
  tile_t *tile, tiledef_t *tileDef, int32_t x, int32_t y, int32_t z
);